# Python standard library
from os import listdir
from random import randrange
from math import pi, sin, cos
from threading import Timer
from typing import List, Tuple

# Dependencies
from psutil import Process 
from dearpygui import dearpygui as dpg

# ft-euler
from ft_euler.dev.TrackImport import get_cones_from_file
from ft_euler.export.RepeatTimer import RepeatTimer
from ft_euler.geometry.Vector2 import Vector2
from ft_euler.pathfinding.Cone import Cone
from ft_euler.pathfinding.Track import Track
from ft_euler.pathfinding.Pathfinder import Pathfinder
from ft_euler.pathfinding.Pose import Pose


class Simulator:
    def __init__(self, track_folder_path: str):
        self.running = True
        self.tracks = {}
        self.selected_track = None
        self.enable_draw_blue = True
        self.enable_connect_blue = False
        self.enable_draw_yellow = True
        self.enable_connect_yellow = False
        self.enable_draw_big_orange = True
        self.enable_draw_orange = True
        self.enable_draw_car_start = True
        self.colorblind = True
        self.car_position = Vector2(0, 0)
        self.car_rotation = 0
        self.enable_visible_radius = False
        self.visible_radius = 5.0
        self.enable_tessellation = True
        self.enable_midpoints = False
        self.enable_ordered_midpoints = True
        self.x_axis = None
        self.y_axis = None
        self.track_folder_path = track_folder_path
        self.cone_x_positions = []
        self.cone_y_positions = []
        self.tessellation_triangles = []
        self.midpoints = []
        self.ordered_midpoints = []
        self.triangle_area_series_list = []
        
        self.refresh_interval = 0.02

        track_filenames = []
        for track_filename in listdir(track_folder_path):
            if track_filename.endswith('.csv'):
                track_filenames.append(track_filename)

        for track_filename in track_filenames:
            key = track_filename[:-4]
            self.tracks[key] = track_filename

        self.process = Process()
        self.ros_timer = None

    def get_cpu_usage(self) -> float:
        cpu_usage = self.process.cpu_percent()
        return cpu_usage

    def update_cpu_usage(self) -> None:
        if self.running:
            dpg.set_value("cpu_usage", "CPU usage: " + str(round(self.get_cpu_usage())) + "%")
            
            # TODO: Add kill switch
            Timer(0.2, self.update_cpu_usage).start()

    def get_memory_usage(self) -> float:
        memory_info = self.process.memory_info()
        memory_usage = memory_info.rss / (1024 ** 2)  # Convert bytes to megabytes
        return memory_usage

    def update_memory_usage(self) -> None:
        if self.running:
            dpg.set_value("memory_usage", "Memory usage: " + str(round(self.get_memory_usage())) + " MB")
            
            # TODO: Add kill switch
            Timer(0.2, self.update_memory_usage).start()

    # Callbacks
    def update_pose_callback(self, car_position: Vector2, car_rotation: float) -> None:
        self.car_position = car_position
        self.car_rotation = car_rotation
        
    def update_pose_callback_with_pose(self, pose: Pose) -> None:
        self.car_position = pose.position
        self.car_rotation = pose.rotation
        self.update_plot()
    
    def import_from_ros_callback(self) -> None:
        print("ROS timer started")
        
        import rclpy
        rclpy.init(args=None)
        
        if self.ros_timer is not None:
            self.ros_timer.cancel()
        
        self.ros_timer = RepeatTimer(self.refresh_interval, self.refresh_ros_data)
        self.ros_timer.setDaemon(True)
        self.ros_timer.start()
        
    def refresh_ros_data(self) -> None:
        from ft_euler.export.RosPlugin import RosConesPlugin, RosPosePlugin
        self.ros_cones_plugin = RosConesPlugin(callback=lambda cones: self.update_track_from_cones("ros", cones))
        self.ros_cones_plugin.start()
        self.ros_pose_plugin = RosPosePlugin(callback=lambda pose: self.update_pose_callback_with_pose(pose))
        self.ros_pose_plugin.start()

    def reset_view_callback(self) -> None:
        self.scale_view()

    def draw_blue_callback(self, sender) -> None:
        self.enable_draw_blue = dpg.get_value(sender)
        self.update_plot()

    def connect_blue_callback(self, sender) -> None:
        self.enable_connect_blue = dpg.get_value(sender)
        self.update_plot()

    def draw_yellow_callback(self, sender) -> None:
        self.enable_draw_yellow = dpg.get_value(sender)
        self.update_plot()

    def connect_yellow_callback(self, sender) -> None:
        self.enable_connect_yellow = dpg.get_value(sender)
        self.update_plot()

    def draw_big_orange_callback(self, sender) -> None:
        self.enable_draw_big_orange = dpg.get_value(sender)
        self.update_plot()

    def draw_orange_callback(self, sender) -> None:
        self.enable_draw_orange = dpg.get_value(sender)
        self.update_plot()

    def draw_car_start_callback(self, sender) -> None:
        self.enable_draw_car_start = dpg.get_value(sender)
        self.update_plot()

    def track_select_callback(self, sender) -> None:
        track_id = dpg.get_value(sender)
        self.update_track(track_id, self.track_folder_path + self.tracks[track_id])

    def resize_viewport_callback(self) -> None:
        dpg.configure_item(self.track_manager,
                           width=(dpg.get_viewport_width() * 0.25),
                           height=(dpg.get_viewport_height()),
                           pos=[0, 0])

        dpg.configure_item(self.track_display, width=(dpg.get_viewport_width() * 0.75),
                           height=(dpg.get_viewport_height()),
                           pos=[dpg.get_viewport_width() * 0.25, 0])

    def update_track(self, track_name: str, track_file_path: str) -> None:
        track_cones = get_cones_from_file(track_file_path)
        self.selected_track = Track(name=track_name, cones=track_cones)
        self.reset_car_position_callback()
        pathfinder = Pathfinder(self.selected_track)
        pathfinder.solve(self.car_position, self.car_rotation, colorblind=self.colorblind)
        self.tessellation_triangles = pathfinder.tessellated_triangles
        self.midpoints = pathfinder.unordered_midpoints
        self.ordered_midpoints = pathfinder.path
        dpg.set_value("cone_count", "Total cones: " + str(len(self.selected_track.cones)))
        dpg.set_value("triangle_count", "Total triangles: " + str(len(self.tessellation_triangles)))
        self.update_plot()
        # self.scale_view()

    # TODO: Make this better
    # For ROS usage
    def update_track_from_cones(self, track_name: str, track_cones: List[Cone]) -> None:
        self.selected_track = Track(name=track_name, cones=track_cones)
        
        # This approach doesn't work with rotation
        for cone in track_cones:
            cone.pos.x += self.car_position.x
            cone.pos.y += self.car_position.y
            
        # self.reset_car_position_callback()
        pathfinder = Pathfinder(self.selected_track)
        pathfinder.solve(self.car_position, self.car_rotation, colorblind=self.colorblind)
        self.tessellation_triangles = pathfinder.tessellated_triangles
        self.midpoints = pathfinder.unordered_midpoints
        self.ordered_midpoints = pathfinder.path
        self.selected_track = pathfinder.track
        dpg.set_value("cone_count", "Total cones: " + str(len(self.selected_track.cones)))
        dpg.set_value("triangle_count", "Total triangles: " + str(len(self.tessellation_triangles)))
        self.update_plot()
        # self.scale_view()

    def reset_car_position_callback(self) -> None:
        print("Reset car position callback")
        if self.selected_track is not None:
            car_start_cone = self.selected_track.get_car_start_cone()
            if car_start_cone is not None:
                self.car_position = car_start_cone.pos
                dpg.set_value("car_position_input", [self.car_position.x, self.car_position.y])
            else:
                self.car_position = Vector2(0, 0)
                dpg.set_value("car_position_input", [self.car_position.x, self.car_position.y])
        else:
            self.car_position = Vector2(0, 0)
            dpg.set_value("car_position_input", [self.car_position.x, self.car_position.y])

        self.update_plot()
        
    def colorblind_callback(self, sender) -> None:
        self.colorblind = dpg.get_value(sender)
        
        # Clear existing triangles
        for triangle_area_series in self.triangle_area_series_list:
            dpg.set_value(triangle_area_series, [[0, 0, 0], [0, 0, 0]])
            
        self.recalculate_path()

    def recalculate_path(self) -> None:
        pathfinder = Pathfinder(self.selected_track)
        pathfinder.solve(self.car_position, self.car_rotation, colorblind=self.colorblind)
        self.tessellation_triangles = pathfinder.tessellated_triangles
        self.midpoints = pathfinder.unordered_midpoints
        self.ordered_midpoints = pathfinder.path
        self.update_plot()

    def car_position_callback(self, sender) -> None:
        # val = dpg.get_value(sender)
        # self.car_position = Vector2(val[0], val[1])
        # self.update_plot()
        pass

    def enable_visible_radius_callback(self, sender) -> None:
        self.enable_visible_radius = dpg.get_value(sender)
        self.update_plot()

    def visible_radius_callback(self, sender) -> None:
        self.visible_radius = dpg.get_value(sender)
        self.update_plot()

    def tessellation_callback(self, sender) -> None:
        self.enable_tessellation = dpg.get_value(sender)
        self.update_plot()

    def midpoints_callback(self, sender) -> None:
        self.enable_midpoints = dpg.get_value(sender)
        self.update_plot()

    def ordered_midpoints_callback(self, sender) -> None:
        self.enable_ordered_midpoints = dpg.get_value(sender)
        self.update_plot()

    # Conversions for plotting
    def triangles_to_plot(self) -> List[Tuple[float, float]]:
        triangles_in_plot_form = []

        for triangle in self.tessellation_triangles:
            triangle_x_positions = []
            triangle_y_positions = []
            for vertex in triangle.points:
                triangle_x_positions.append(vertex.x)
                triangle_y_positions.append(vertex.y)
            triangle = [triangle_x_positions, triangle_y_positions]
            triangles_in_plot_form.append(triangle)

        return triangles_in_plot_form

    def blue_cones_to_plot(self) -> Tuple[list, list]:
        x_positions = []
        y_positions = []

        if self.selected_track is not None:
            for cone in self.selected_track.get_blue_cones():
                x_positions.append(cone.pos.x)
                y_positions.append(cone.pos.y)

        return x_positions, y_positions

    def yellow_cones_to_plot(self) -> Tuple[list, list]:
        x_positions = []
        y_positions = []

        if self.selected_track is not None:
            for cone in self.selected_track.get_yellow_cones():
                x_positions.append(cone.pos.x)
                y_positions.append(cone.pos.y)

        return x_positions, y_positions

    def big_orange_cones_to_plot(self) -> Tuple[list, list]:
        x_positions = []
        y_positions = []

        if self.selected_track is not None:
            for cone in self.selected_track.cones:
                if cone.color == "big_orange":
                    x_positions.append(cone.pos.x)
                    y_positions.append(cone.pos.y)

        return x_positions, y_positions

    def orange_cones_to_plot(self) -> Tuple[list, list]:
        x_positions = []
        y_positions = []

        if self.selected_track is not None:
            for cone in self.selected_track.cones:
                if cone.color == "orange":
                    x_positions.append(cone.pos.x)
                    y_positions.append(cone.pos.y)

        return x_positions, y_positions

    def car_start_to_plot(self) -> Tuple[list, list]:
        x_positions = []
        y_positions = []

        if self.selected_track is not None:
            for cone in self.selected_track.cones:
                if cone.color == "car_start":
                    x_positions.append(cone.pos.x)
                    y_positions.append(cone.pos.y)

        return x_positions, y_positions

    def midpoints_to_plot(self) -> Tuple[list, list]:
        x_positions = []
        y_positions = []

        for midpoint in self.midpoints:
            x_positions.append(midpoint.x)
            y_positions.append(midpoint.y)

        return x_positions, y_positions

    def ordered_midpoints_to_plot(self) -> Tuple[list, list]:
        x_positions = []
        y_positions = []

        for ordered_midpoint in self.ordered_midpoints:
            x_positions.append(ordered_midpoint.x)
            y_positions.append(ordered_midpoint.y)

        return x_positions, y_positions

    def visible_radius_to_plot(self) -> Tuple[list, list]:
        x_positions = []
        y_positions = []
        num_points = 64
        angle_increment = 2 * pi / num_points

        for i in range(num_points - 1):
            angle = i * angle_increment
            x = self.car_position.x + (self.visible_radius * cos(angle))
            y = self.car_position.y + (self.visible_radius * sin(angle))
            x_positions.append(x)
            y_positions.append(y)

        # Close the loop
        x_positions.append(x_positions[0])
        y_positions.append(y_positions[0])

        return x_positions, y_positions

    def car_position_to_plot(self) -> List[List[int]]:
        return [[self.car_position.x], [self.car_position.y]]

    # Plot rendering
    def update_plot(self) -> None:
        if self.enable_tessellation:
            # for triangle_area_series in self.triangle_area_series_list:
            #     dpg.set_value(triangle_area_series, [[0, 0, 0], [0, 0, 0]])

            triangles = self.triangles_to_plot()
            for index, triangle in enumerate(triangles):
                dpg.set_value(self.triangle_area_series_list[index], triangle)
        else:
            for triangle_area_series in self.triangle_area_series_list:
                dpg.set_value(triangle_area_series, [[0, 0, 0], [0, 0, 0]])

        if self.enable_visible_radius and self.selected_track is not None:
            dpg.set_value("visible_radius_plot", self.visible_radius_to_plot())
        else:
            dpg.set_value("visible_radius_plot", [list(), list()])

        if len(self.ordered_midpoints) > 0 and self.enable_ordered_midpoints:
            ordered_midpoints = self.ordered_midpoints_to_plot()
            # Add first point to end to complete the loop
            # ordered_midpoints[0].append(ordered_midpoints[0][0])
            # ordered_midpoints[1].append(ordered_midpoints[1][0])
            dpg.set_value("ordered_midpoints_line_series", ordered_midpoints)
        else:
            dpg.set_value("ordered_midpoints_line_series", [list(), list()])

        if self.selected_track is not None and self.enable_connect_blue:
            blue_cones = self.blue_cones_to_plot()
            # Add first point to end to complete the loop
            blue_cones[0].append(blue_cones[0][0])
            blue_cones[1].append(blue_cones[1][0])
            dpg.set_value('blue_line_series', blue_cones)
        else:
            dpg.set_value('blue_line_series', [list(), list()])

        if self.selected_track is not None and self.enable_connect_yellow:
            yellow_cones = self.yellow_cones_to_plot()
            # Add first point to end to complete the loop
            yellow_cones[0].append(yellow_cones[0][0])
            yellow_cones[1].append(yellow_cones[1][0])
            dpg.set_value('yellow_line_series', yellow_cones)
        else:
            dpg.set_value('yellow_line_series', [list(), list()])

        if self.enable_draw_blue:
            dpg.set_value('blue_cone_plot', self.blue_cones_to_plot())
        else:
            dpg.set_value('blue_cone_plot', [list(), list()])

        if self.enable_draw_yellow:
            dpg.set_value('yellow_cone_plot', self.yellow_cones_to_plot())
        else:
            dpg.set_value('yellow_cone_plot', [list(), list()])

        if self.enable_draw_big_orange:
            dpg.set_value('big_orange_cone_plot', self.big_orange_cones_to_plot())
        else:
            dpg.set_value('big_orange_cone_plot', [list(), list()])

        if self.enable_draw_orange:
            dpg.set_value('orange_cone_plot', self.orange_cones_to_plot())
        else:
            dpg.set_value('orange_cone_plot', [list(), list()])

        if self.enable_draw_car_start:
            dpg.set_value('car_start_cone_plot', self.car_start_to_plot())
        else:
            dpg.set_value('car_start_cone_plot', [list(), list()])

        if self.enable_midpoints:
            dpg.set_value('midpoints_plot', self.midpoints_to_plot())
        else:
            dpg.set_value('midpoints_plot', [list(), list()])

        dpg.set_value("car_position_plot", self.car_position_to_plot())

    def scale_view(self) -> None:
        dpg.fit_axis_data(self.x_axis)
        dpg.fit_axis_data(self.y_axis)

    def start(self) -> None:
        dpg.create_context()
        dpg.create_viewport(title="ft-euler", resizable=True, width=1200, height=600)
        dpg.set_viewport_resize_callback(self.resize_viewport_callback)
        dpg.setup_dearpygui()

        # blue_cone_color = (0, 0, 255)
        # yellow_cone_color = (255, 255, 0)
        # big_orange_cone_color = (139, 64, 0)
        # orange_cone_color = (255, 165, 0)
        # car_start_cone_color = (255, 0, 0)

        blue_cone_color = (255, 255, 255)
        yellow_cone_color = (255, 255, 255)
        big_orange_cone_color = (255, 255, 255)
        orange_cone_color = (255, 255, 255)
        car_start_cone_color = (255, 255, 255)

        with dpg.window(label="Track manager", no_close=True, width=int(dpg.get_viewport_width() * 0.25), height=dpg.get_viewport_height(), no_collapse=True,
                        no_resize=True, no_move=True, tag="track_manager") as self.track_manager:
            dpg.add_separator()
            dpg.add_text("Tracks")
            dpg.add_button(label="Import from ROS (requires rospy)", callback=self.import_from_ros_callback)
            dpg.add_combo(default_value="Select a track", items=list(self.tracks.keys()),
                          callback=self.track_select_callback, width=250)
            dpg.add_button(label="Reset view", callback=self.reset_view_callback)
            dpg.add_separator()

            dpg.add_text("Cones")
            dpg.add_checkbox(default_value=self.enable_draw_blue, label="Blue cones", callback=self.draw_blue_callback)
            dpg.add_checkbox(show=False, default_value=self.enable_connect_blue, label="Connect blue cones (WIP)",
                             callback=self.connect_blue_callback)
            dpg.add_checkbox(default_value=self.enable_draw_yellow, label="Yellow cones",
                             callback=self.draw_yellow_callback)
            dpg.add_checkbox(show=False, default_value=self.enable_connect_yellow, label="Connect yellow cones (WIP)",
                             callback=self.connect_yellow_callback)
            dpg.add_checkbox(default_value=self.enable_draw_big_orange, label="Big orange cones",
                             callback=self.draw_big_orange_callback)
            dpg.add_checkbox(default_value=self.enable_draw_orange, label="Orange cones",
                             callback=self.draw_orange_callback)
            dpg.add_checkbox(default_value=self.enable_draw_car_start, label="Car start cone",
                             callback=self.draw_car_start_callback)
            dpg.add_separator()

            dpg.add_text("Pathfinding")
            dpg.add_checkbox(default_value=self.colorblind, label="Colorblind", callback=self.colorblind_callback)
            dpg.add_button(show=False, label="Recalculate path", callback=self.recalculate_path)
            dpg.add_button(show=False, label="Reset car position", callback=None)
            dpg.add_slider_floatx(show=False, default_value=[self.car_position.x, self.car_position.y], label="Car position",
                                 callback=None, size=2, min_value=-50, max_value=50,
                                 tag="car_position_input")
            dpg.add_checkbox(show=False, default_value=self.enable_visible_radius, label="Visible radius (m) (WIP)",
                             callback=self.enable_visible_radius_callback)
            dpg.add_slider_float(show = False, default_value=self.visible_radius,
                                 callback=self.visible_radius_callback, min_value=0.25, max_value=15)
            dpg.add_checkbox(default_value=self.enable_tessellation, label="Tessellation",
                             callback=self.tessellation_callback)
            dpg.add_checkbox(default_value=self.enable_midpoints, label="Midpoints",
                             callback=self.midpoints_callback)
            dpg.add_checkbox(default_value=self.enable_ordered_midpoints, label="Path",
                             callback=self.ordered_midpoints_callback)
            dpg.add_separator()

            dpg.add_text("Statistics")
            dpg.add_text(default_value="Total cones: 0", label="Cone count", tag="cone_count")
            dpg.add_text(default_value="Total triangles: 0", label="Triangle count (tessellation)",
                         tag="triangle_count")

            dpg.add_separator()

            dpg.add_text("Performance")
            dpg.add_text(default_value="CPU usage: 0%", label="CPU usage", tag="cpu_usage")
            dpg.add_text(default_value="Memory usage: 0 MB", label="Memory usage", tag="memory_usage")
            dpg.add_separator()

        with dpg.window(label="Track display", pos=[dpg.get_viewport_width() * 0.25, 0], width=int(dpg.get_viewport_width() * 0.75), height=dpg.get_viewport_height(), no_close=True, no_collapse=True,
                        no_move=True, no_resize=True) as self.track_display:
            with dpg.theme(tag="blue_theme"):
                with dpg.theme_component(dpg.mvScatterSeries):
                    dpg.add_theme_color(dpg.mvPlotCol_Line, blue_cone_color, category=dpg.mvThemeCat_Plots)
                    dpg.add_theme_style(dpg.mvPlotStyleVar_Marker, dpg.mvPlotMarker_Circle,
                                        category=dpg.mvThemeCat_Plots)
                    dpg.add_theme_style(dpg.mvPlotStyleVar_MarkerSize, 5, category=dpg.mvThemeCat_Plots)

            with dpg.theme(tag="yellow_theme"):
                with dpg.theme_component(dpg.mvScatterSeries):
                    dpg.add_theme_color(dpg.mvPlotCol_Line, yellow_cone_color, category=dpg.mvThemeCat_Plots)
                    dpg.add_theme_style(dpg.mvPlotStyleVar_Marker, dpg.mvPlotMarker_Circle,
                                        category=dpg.mvThemeCat_Plots)
                    dpg.add_theme_style(dpg.mvPlotStyleVar_MarkerSize, 5, category=dpg.mvThemeCat_Plots)

            with dpg.theme(tag="big_orange_theme"):
                with dpg.theme_component(dpg.mvScatterSeries):
                    dpg.add_theme_color(dpg.mvPlotCol_Line, big_orange_cone_color, category=dpg.mvThemeCat_Plots)
                    dpg.add_theme_style(dpg.mvPlotStyleVar_Marker, dpg.mvPlotMarker_Circle,
                                        category=dpg.mvThemeCat_Plots)
                    dpg.add_theme_style(dpg.mvPlotStyleVar_MarkerSize, 5, category=dpg.mvThemeCat_Plots)

            with dpg.theme(tag="orange_theme"):
                with dpg.theme_component(dpg.mvScatterSeries):
                    dpg.add_theme_color(dpg.mvPlotCol_Line, orange_cone_color, category=dpg.mvThemeCat_Plots)
                    dpg.add_theme_style(dpg.mvPlotStyleVar_Marker, dpg.mvPlotMarker_Circle,
                                        category=dpg.mvThemeCat_Plots)
                    dpg.add_theme_style(dpg.mvPlotStyleVar_MarkerSize, 5, category=dpg.mvThemeCat_Plots)

            with dpg.theme(tag="car_start_theme"):
                with dpg.theme_component(dpg.mvScatterSeries):
                    dpg.add_theme_color(dpg.mvPlotCol_Line, car_start_cone_color, category=dpg.mvThemeCat_Plots)
                    dpg.add_theme_style(dpg.mvPlotStyleVar_Marker, dpg.mvPlotMarker_Circle,
                                        category=dpg.mvThemeCat_Plots)
                    dpg.add_theme_style(dpg.mvPlotStyleVar_MarkerSize, 5, category=dpg.mvThemeCat_Plots)

            with dpg.theme(tag="midpoints_theme"):
                with dpg.theme_component(dpg.mvScatterSeries):
                    dpg.add_theme_color(dpg.mvPlotCol_Line, (0, 255, 0, 255), category=dpg.mvThemeCat_Plots)
                    dpg.add_theme_style(dpg.mvPlotStyleVar_Marker, dpg.mvPlotMarker_Plus,
                                        category=dpg.mvThemeCat_Plots)
                    dpg.add_theme_style(dpg.mvPlotStyleVar_MarkerSize, 5, category=dpg.mvThemeCat_Plots)
                    dpg.add_theme_style(dpg.mvPlotStyleVar_MarkerWeight, 3, category=dpg.mvThemeCat_Plots)

            with dpg.theme(tag="ordered_midpoints_theme"):
                with dpg.theme_component(dpg.mvLineSeries):
                    dpg.add_theme_style(dpg.mvPlotStyleVar_LineWeight, 6, category=dpg.mvThemeCat_Plots)

            with dpg.theme(tag="visible_radius_theme"):
                with dpg.theme_component(dpg.mvAreaSeries):
                    dpg.add_theme_color(dpg.mvPlotCol_Fill, (255, 255, 255, 50), category=dpg.mvThemeCat_Plots)
                    # dpg.add_theme_style(dpg.mvPlotStyleVar_MarkerSize, self.visible_radius,
                    #                     category=dpg.mvThemeCat_Plots, tag="visible_radius_plot_size")
                    pass

            with dpg.theme(tag="car_position_theme"):
                with dpg.theme_component(dpg.mvScatterSeries):
                    dpg.add_theme_color(dpg.mvPlotCol_Line, (255, 255, 255, 255), category=dpg.mvThemeCat_Plots)
                    dpg.add_theme_style(dpg.mvPlotStyleVar_Marker, dpg.mvPlotMarker_Square,
                                        category=dpg.mvThemeCat_Plots)
                    dpg.add_theme_style(dpg.mvPlotStyleVar_MarkerSize, 8, category=dpg.mvThemeCat_Plots)
                    dpg.add_theme_style(dpg.mvPlotStyleVar_MarkerWeight, 3, category=dpg.mvThemeCat_Plots)

            with dpg.plot(label="", no_box_select=True, equal_aspects=True, tag="track_plot") as self.track_plot:
                self.x_axis = dpg.add_plot_axis(dpg.mvXAxis, label="x (m)")
                self.y_axis = dpg.add_plot_axis(dpg.mvYAxis, label="y (m)")

                for i in range(1000):
                    self.triangle_area_series_list.append(dpg.add_area_series([0, 0, 0], [0, 0, 0],
                                                                              fill=[randrange(100, 255),
                                                                                    randrange(100, 255),
                                                                                    randrange(100, 255), 128],
                                                                              parent=self.y_axis))

                dpg.add_line_series(self.cone_x_positions, self.cone_y_positions,
                                    tag='blue_line_series', parent=self.y_axis)

                dpg.add_line_series(self.cone_x_positions, self.cone_y_positions,
                                    tag='yellow_line_series', parent=self.y_axis)

                ordered_midpoints_line_series = dpg.add_line_series(self.cone_x_positions, self.cone_y_positions,
                                                                    tag='ordered_midpoints_line_series',
                                                                    parent=self.y_axis)

                dpg.bind_item_theme(ordered_midpoints_line_series, "ordered_midpoints_theme")

                visible_radius_area_series = dpg.add_area_series(self.cone_x_positions, self.cone_y_positions,
                                                                 tag='visible_radius_plot',
                                                                 parent=self.y_axis)

                dpg.bind_item_theme(visible_radius_area_series, 'visible_radius_theme')

                blue_cone_scatter_series = dpg.add_scatter_series(self.cone_x_positions, self.cone_y_positions,
                                                                  tag='blue_cone_plot',
                                                                  parent=self.y_axis)

                dpg.bind_item_theme(blue_cone_scatter_series, "blue_theme")

                yellow_cone_scatter_series = dpg.add_scatter_series(self.cone_x_positions, self.cone_y_positions,
                                                                    tag='yellow_cone_plot',
                                                                    parent=self.y_axis)

                dpg.bind_item_theme(yellow_cone_scatter_series, "yellow_theme")

                big_orange_cone_scatter_series = dpg.add_scatter_series(self.cone_x_positions, self.cone_y_positions,
                                                                        tag='big_orange_cone_plot',
                                                                        parent=self.y_axis)

                dpg.bind_item_theme(big_orange_cone_scatter_series, "big_orange_theme")

                orange_cone_scatter_series = dpg.add_scatter_series(self.cone_x_positions, self.cone_y_positions,
                                                                    tag='orange_cone_plot',
                                                                    parent=self.y_axis)

                dpg.bind_item_theme(orange_cone_scatter_series, "orange_theme")

                car_start_cone_scatter_series = dpg.add_scatter_series(self.cone_x_positions, self.cone_y_positions,
                                                                       tag='car_start_cone_plot',
                                                                       parent=self.y_axis)

                dpg.bind_item_theme(car_start_cone_scatter_series, "car_start_theme")

                midpoint_scatter_series = dpg.add_scatter_series(self.cone_x_positions, self.cone_y_positions,
                                                                 tag='midpoints_plot',
                                                                 parent=self.y_axis)

                dpg.bind_item_theme(midpoint_scatter_series, "midpoints_theme")

                car_position_scatter_series = dpg.add_scatter_series(self.cone_x_positions, self.cone_y_positions,
                                                                     tag="car_position_plot",
                                                                     parent=self.y_axis)

                dpg.bind_item_theme(car_position_scatter_series, "car_position_theme")

        dpg.show_viewport()

        self.update_cpu_usage()
        self.update_memory_usage()

        while dpg.is_dearpygui_running():
            dpg.configure_item(self.track_plot, width=(dpg.get_item_width(self.track_display) - 20))
            dpg.configure_item(self.track_plot, height=(dpg.get_item_height(self.track_display) - 20))
            dpg.render_dearpygui_frame()

        self.running = False
        dpg.destroy_context()


def main() -> None:
    sim = Simulator("../../tracks/")
    sim.start()


if __name__ == "__main__":
    main()
