# Python standard library
from random import uniform
from typing import Union
from time import time

# Dependencies
from rclpy import node, spin_once
from eufs_msgs.msg import (ConeArrayWithCovariance, PointArray,
                           FullState, WaypointArrayStamped, Waypoint, CarState)
from geometry_msgs.msg import Point, Pose, PoseStamped, Quaternion
from nav_msgs.msg import Path
from scipy import interpolate
from std_msgs.msg import Header, Bool
import numpy as np

# ft-euler
from ft_euler.geometry.Utilities import quaternion_to_yaw
from ft_euler.pathfinding.Cone import Cone
from ft_euler.pathfinding.Track import Track
from ft_euler.pathfinding.Triangulation import *
from ft_euler.pathfinding.Pathfinder import Pathfinder
from ft_euler.pathfinding.Pose import Pose as euler_Pose

def cone_array_with_covariance_to_internal_cone_format(cones, max_noise=0.0, swap_color_prob=0.0):
    inner_cones = []
    
    for cone in cones.yellow_cones:
        inner_cones.append(Cone('yellow', cone.point.x + uniform(-max_noise, max_noise), cone.point.y +
                                uniform(-max_noise, max_noise)))
    for cone in cones.orange_cones:
        inner_cones.append(Cone('orange', cone.point.x + uniform(-max_noise, max_noise), cone.point.y +
                                uniform(-max_noise, max_noise)))
    for cone in cones.blue_cones:
        inner_cones.append(Cone('blue', cone.point.x + uniform(-max_noise, max_noise), cone.point.y +
                                uniform(-max_noise, max_noise)))
    for cone in cones.orange_cones:
        inner_cones.append(Cone('orange', cone.point.x + uniform(-max_noise, max_noise), cone.point.y +
                                uniform(-max_noise, max_noise)))
    for cone in cones.big_orange_cones:
        inner_cones.append(Cone('big_orange', cone.point.x + uniform(-max_noise, max_noise), cone.point.y +
                                uniform(-max_noise, max_noise)))

    return inner_cones

POINT_DENSITY = 100

def full_state_to_internal_pose_format(full_state: FullState) -> euler_Pose:
    position = Vector2(full_state.x_pos, full_state.y_pos)
    pose = euler_Pose(position, full_state.yaw)
    return pose

def car_state_to_internal_pose_format(car_state: CarState) -> euler_Pose:
    position = Vector2(car_state.pose.pose.position.x, car_state.pose.pose.position.y)
    rotation = quaternion_to_yaw(car_state.pose.pose.orientation)
    pose = euler_Pose(position, rotation)
    return pose

def pose_stamped_to_internal_pose_format(pose_stamped: PoseStamped) -> euler_Pose:
    position = Vector2(pose_stamped.pose.position.x, pose_stamped.pose.position.y)
    rotation = quaternion_to_yaw(pose_stamped.pose.orientation)
    pose = euler_Pose(position, rotation)
    return pose
        
class RosConesPlugin:
    node = None
    
    def __init__(self, callback):
        if not RosConesPlugin.node:
            RosConesPlugin.node = node.Node("ft_euler_ros_cones_plugin_node")
            
        self.cone_subscription = RosConesPlugin.node.create_subscription(ConeArrayWithCovariance, '/map', self.callback, 1)

        self.outer_callback = callback
        self.bad_spline_num = 0
        
    def start(self):
        # print("Updating RosConesPlugin")
        spin_once(RosConesPlugin.node)
        
    def callback(self, cones):
        cones = cone_array_with_covariance_to_internal_cone_format(cones)
        self.outer_callback(cones)
        RosConesPlugin.node.destroy_subscription(self.cone_subscription)

class RosPosePlugin:
    node = None
    
    def __init__(self, callback):
        if not RosPosePlugin.node:
            RosPosePlugin.node = node.Node("ft_euler_ros_pose_plugin_node")
            
        # self.pose_subscription = RosPosePlugin.node.create_subscription(FullState, '/pose', self.callback, 1)
        # self.pose_subscription = RosPosePlugin.node.create_subscription(CarState, '/odometry_integration/car_state', self.callback, 1)
        self.pose_subscription = RosPosePlugin.node.create_subscription(PoseStamped, '/pose', self.callback, 1)
        self.outer_callback = callback
        
    def start(self):
        # print("Updating RosPosePlugin")
        spin_once(RosPosePlugin.node)
        
    def callback(self, pose: Union[FullState, CarState, PoseStamped]):
        if type(pose) == FullState:
            pose = full_state_to_internal_pose_format(pose)
        elif type(pose) == CarState:
            pose = car_state_to_internal_pose_format(pose)
        elif type(pose) == PoseStamped:
            pose = pose_stamped_to_internal_pose_format(pose)
        else:
            print("Invalid pose input of type " + str(type(pose)))
            
        self.outer_callback(pose)
        RosPosePlugin.node.destroy_subscription(self.pose_subscription)

class Node(node.Node):
    
    def __init__(self):
        super().__init__("ft_euler_node")
        self.cones_subscription = self.create_subscription(ConeArrayWithCovariance, "/map", self.cones_callback, 1)
        self.pose_subscription = self.create_subscription(PoseStamped, "/pose", self.pose_callback, 1)
        self.raceline_publisher = self.create_publisher(PointArray, "/raceline", 1)
        self.waypoint_publisher = self.create_publisher(WaypointArrayStamped, "/trajectory", 1)
        self.rviz_path_publisher = self.create_publisher(Path, "/path", 1)
        self.rotated_publisher = self.create_publisher(WaypointArrayStamped, "/rotated", 1)
        self.change_path = self.create_publisher(Bool, "/change_path", 1)
        self.last_solve_time = 0.0

        self.pose = Vector2(0, 0)
        self.points = [Point(x=float(val)*0.1, y=0.0, z=0.0) for val in np.arange(1, 100)] # Extended line 10m ahead of the car
        self.pose_rot= np.zeros((3, 1))

    def pose_callback(self, pose):
        yaw = quaternion_to_yaw(pose.pose.orientation)
        self.pose = Vector2(pose.pose.position.x, pose.pose.position.y)
        self.pose_rot = np.array([[pose.pose.position.x], \
                                  [pose.pose.position.y], \
                                  [yaw]])
        # TODO : Change when at comp
                            
    def cones_callback(self, cones):
        cones = cone_array_with_covariance_to_internal_cone_format(cones)
        track = Track(name="ros", cones=cones)
        pathfinder = Pathfinder(track)
        
        # rotate points function
        # check curviness, return bool function
        
        if time() - self.last_solve_time > 0.1:
            pathfinder.solve(self.pose, self.pose_rot[2], colorblind=True)
            self.last_solve_time = time()
            
            # START ROTATION
            points = np.array([(point.x, point.y) for point in pathfinder.path])
            unique=[]
            # iterating each array of array
            
            for i in points:
                # if array is not present in the list
                # add the array to list.
                if list(i) not in unique:
                    unique.append(list(i))
            points = np.array(unique)
            
            rot_matrix = np.array([[np.cos(self.pose_rot[2, 0]), -np.sin(self.pose_rot[2, 0])], [np.sin(self.pose_rot[2, 0]), np.cos(self.pose_rot[2, 0])]]) 
            rot_points = [point.reshape(2,1) - self.pose_rot[:2] for point in points]
            rot_points = np.array([rot_matrix@point for point in rot_points])

            scrap_points = []
            for idx, rot in enumerate(rot_points):
                # [front x max, front y max, front y min, back y min, back y max, back y min]
                bounds = [10, 8, -8, -5, 5, -5]

                # yaw_tol = np.pi / 4
                # x_rad, y_rad = self.get_radius(length=30)
                # sector_bounds = [30*np.cos(self.pose_rot[2,0] - yaw_tol), 
                                #  30*np.sin(self.pose_rot[2,0] - yaw_tol),
                                #  30*np.cos(self.pose_rot[2,0] + yaw_tol),
                                #  30*np.sin(self.pose_rot[2,0] + yaw_tol)]

                if rot[0] >= 0: #Front box
                    if rot[0] > bounds[0]:
                        scrap_points.append(idx)    
                    if rot[1] > bounds[1]:
                        scrap_points.append(idx)
                    if rot[1] < bounds[2]: 
                        scrap_points.append(idx)
                else: 
                    if rot[0] < bounds[3]:
                        scrap_points.append(idx)    
                    if rot[1] > bounds[4]:
                        scrap_points.append(idx)
                    if rot[1] < bounds[5]: 
                        scrap_points.append(idx)

            rot_points = np.delete(rot_points, scrap_points, axis=0)
            
            points = np.delete(points, scrap_points, axis=0)
            rot_waypoints = WaypointArrayStamped(header=Header(frame_id="base_footprint"))
            for i in points:
                rotpoint = Point()
                rotpoint.x, rotpoint.y = float(i[0]), float(i[1])
                waypoint_rot = Waypoint()
                waypoint_rot.position = rotpoint
                rot_waypoints.waypoints.append(waypoint_rot)
            self.rotated_publisher.publish(rot_waypoints)

            if len(rot_points) > 0:
                select = np.argmin(rot_points[:, 0])
                point_select = points[select]
                points_ordered = [point_select]
                while len(points) > 1:
                    points = np.delete(points, select, axis=0)
                    dists = np.array([np.sqrt(np.power(points[idx][0] - point_select[0], 2) + np.power(points[idx][1] - point_select[1], 2)) for idx in range(len(points))])
                    select = np.argmin(dists)
                    point_select = points[select]

                    points_ordered.append(point_select)  
                points_select = points_ordered[0:9] if len(points_ordered) > 10 else points_ordered # TODO Check if 0:5 or 0:4, i think it's 0:5

                points_select = self.make_spline(np.array(points_select))

                curviness = self.check_curviness(points_select)

                # If there is no valid spline then use previous spline
                # However if there is no valid spline for x repetitions
                # then just use raw points for pure pursuit
                
                if curviness:
                    self.points = [Point(x=point.x, y=point.y, z=0.0) for point in points_select]
                    self.bad_spline_num = 0 #Rest counter, good spline
                    msg = Bool()
                    msg.data = False
                    self.change_path.publish(msg)
                else:
                    self.bad_spline_num += 1
                    if self.bad_spline_num > 3:
                        msg = Bool()
                        msg.data = True
                        self.change_path.publish(msg)
                    #send a flag to pure pursuit to use raw points
        
            else:
                #self.get_logger().info("len(rot_points) == 0")
                self.bad_spline_num += 1
                if self.bad_spline_num > 3:
                    msg = Bool()
                    msg.data = True
                    self.change_path.publish(msg)
                    #send a flag to pure pursuit to use raw points
                
                ### self.points remains the same as the last iteration


        self.raceline_publisher.publish(PointArray(points=self.points))
        self.waypoint_publisher.publish(WaypointArrayStamped(header=Header(frame_id="chassis"),
                                                             waypoints=[Waypoint(position=point) for point in self.points]))

        self.rviz_path_publisher.publish(Path(header=Header(stamp=self.get_clock().now().to_msg(),frame_id="base_footprint"),
                                                poses=[PoseStamped(pose=Pose(position=point, orientation=Quaternion(x=0.0,y=0.0,z=0.0,w=1.0))) for point in self.points]))

    def check_curviness(self, points_select):
        # Calculate curvature value reflecting the average change in slope across points in points_select
        # points_select passed into function
        # Trust the first spline because you cant use the previous one?
        # Initialise
        
        total_curve = 0.0
        loop_num = 1

        max_possible_curvature = 2.4 
        max_local_curvature = 14	

        prev_point = Point

        prev_slope = 0.0
        local_truth = False
        for current_point in points_select:
            if loop_num > 1:
                slope = (current_point.y - prev_point.y) / (current_point.x - prev_point.x)
                if loop_num > 2:
                    # Needed absolute value to account for positive and negative valid slopes
                    local_curve = abs(slope - prev_slope)
                    if local_curve > max_local_curvature:
                        local_truth = True 
                        
                    #counter for total difference/curviness
                    total_curve += local_curve
                prev_slope = slope
            prev_point = current_point
            loop_num += 1
        # Find average curvature
        curvature = total_curve / loop_num
        # True means valid spline and false means invalid

        #self.get_logger().info(str("Average_curve: " + str(curvature)))
        if curvature > max_possible_curvature or local_truth:
            #self.get_logger().info("NOT RIGHT PATH")  
            return False
        else:
            #self.get_logger().info("RIGHT PATH")
            return True

    def make_spline(self, points_select):
        if len(points_select) >= 6:
            smooth = self.B_spline(points_select[:, 0], points_select[:,1])
            Xs,Ys = smooth
        elif len(points_select) >= 2:
            Xs, Ys = self.linear_interpolate(points_select[:, 0], points_select[:, 1])

        if len(points_select) >= 2:
            return [Point(x=Xs[idx], y=Ys[idx]) for idx in range(len(Xs))]
        else:
            return []

    def linear_interpolate(self, x, y):
        m, c = np.polyfit(x, y, 1)
        xs = np.linspace(x[0], x[-1], num=POINT_DENSITY)
        ys = m*xs + c

        return xs, ys

    def B_spline(self, x, y, point_density=POINT_DENSITY):
        # spline preperation function (splrep) is being run. We can specify the order of the spline and the smoothness of the representation.
        # The order : k = 3 (default)
        # The smoothness : s = 0 (default)
        # The output of this method is a tuple called tck. It contains the spline knots - coefficients - degree
        
        k = 3

        if len(x) <= 2:
            return self.linear_interpolate(x, y)
        if len(x) <= 4:
            k = 2
        
        tck, *rest = interpolate.splprep([x, y], k=k, s=0)
        # How many values between zero and one you want to take. 
        # We are using 100 in this example but this can be changed.
        u = np.linspace(0., 0.9, num=point_density)
        # This uses a spline evaluation function (splev)
        # smooth = interpolate.splev(u, tck)
        smooth=interpolate.splev(u, tck)
        return smooth

    def get_radius(self, length=1):
        norm = np.sqrt(np.power(self.pose_rot[0,0], 2) + np.power(self.pose_rot[1,0], 2))
        x_len = (length * self.pose_rot[0,0]) / norm
        y_len = (length * self.pose_rot[1,0]) / norm
                
        return x_len, y_len
    
    def areClockwise(self, v1, v2):
        return -v1[0]*v2[1] + v1[1]*v2[0] > 0
