# Python standard library
from math import atan
from typing import Union

# Dependencies
from typing_extensions import Self

# ft-euler
from ft_euler.geometry.Vector2 import Vector2


class Line:
    def __init__(self, point1: Vector2, point2: Vector2) -> None:
        self.p1 = point1
        self.p2 = point2

    def __str__(self) -> None:
        return str(self.p1) + ", " + str(self.p2)

    def calculate_length(self) -> float:
        return Vector2.distance(self.p1, self.p2)

    def calculate_slope(self) -> float:
        # Slope (m) = y2-y1/x2-x1
        if self.p1.x != self.p2.x:
            slope = (self.p2.y - self.p1.y) / (self.p2.x - self.p1.x)
        else:
            slope = float("inf")

        return slope

    def calculate_y_intercept(self) -> Union[float, None]:
        # y-intercept (b) = y1 - m(x1)
        if (self.calculate_slope() == float("inf")):
            return None                                     # Vertical lines don't have a y-intercept
        else:
            return self.p1.y - self.calculate_slope() * self.p1.x

    @staticmethod
    def angle_between_two_lines(l1: Self, l2: Self) -> Union[float, None]:
        l1_slope = l1.calculate_slope()
        l2_slope = l2.calculate_slope()
        
        if (Line.intersect(l1, l2) == False):
            return None                                     # Parallel lines/ lines that never intersect have no angle between them
        elif (l1 == l2):
            return float(0)
        else:
            return atan(abs((l1_slope - l2_slope) / (1 + (l1_slope * l2_slope))))

    @staticmethod
    def point_of_intersection(l1: Self, l2: Self) -> Vector2:
        x_intersect, y_intersect = None, None

        l1_slope = l1.calculate_slope()
        l2_slope = l2.calculate_slope()

        l1_y_intercept = l1.calculate_y_intercept()
        l2_y_intercept = l2.calculate_y_intercept()

        # Parallel lines with different y-intercepts will never intersect
        if l1_slope == l2_slope and l1_y_intercept != l2_y_intercept:
            return Vector2(x_intersect, y_intersect)

        # Avoid dividing by zero
        if l1_slope != l2_slope:
            x_intersect = (l2_y_intercept - l1_y_intercept) / (l1_slope - l2_slope)
        else:
            return Vector2(x_intersect, y_intersect)

        # The point of intersection must be within x-range of both lines
        if x_intersect is not None and (
                x_intersect < min(l1.p1.x, l1.p2.x) or
                x_intersect > max(l1.p1.x, l1.p2.x) or
                x_intersect < min(l2.p1.x, l2.p2.x) or
                x_intersect > max(l2.p1.x, l2.p2.x)):
            return Vector2(x_intersect, y_intersect)

        # y = mx + b
        y_intersect = l1_slope * x_intersect + l1_y_intercept

        return Vector2(x_intersect, y_intersect)

    @staticmethod
    def intersect(l1: Self, l2: Self) -> bool:
        point_of_intersection = Line.point_of_intersection(l1, l2)
        if point_of_intersection.x is not None and point_of_intersection.y is not None:
            return True
        else:
            return False

    def __eq__(self, other) -> bool:
        if self.p1 == other.p1 and self.p2 == other.p2:
            return True
        if self.p2 == other.p1 and self.p1 == other.p2:
            return True

        return False
