# Python standard library
from time import time_ns
from math import pi, sin, cos
from typing import List

# Dependencies
import numpy as np

# ft-euler
from ft_euler.geometry.Line import Line
from ft_euler.geometry.Triangle import Triangle
from ft_euler.geometry.Utilities import point_in_circle, point_is_on_left_side_of_line
from ft_euler.geometry.Vector2 import Vector2
from ft_euler.pathfinding.Cone import Cone
from ft_euler.pathfinding.Track import Track
from ft_euler.pathfinding.Triangulation import tessellate_track


class Pathfinder:    
    def __init__(self, track: Track) -> None:
        self.track = track
        self.tessellated_triangles = None
        self.unordered_midpoints = None
        self.path = None
        
    def intersects_with_track_cones(self, point: Vector2, cone_radius: float) -> bool:
        """Checks if a given point intrsects with any cones on the track."""
        for cone in self.track.cones:
            if point_in_circle(cone.pos, point, cone_radius):
                return True
        
        return False
        
    def get_dividing_line(self, start_position: Vector2, start_rotation: float) -> Line:
        """Finds the best line to divide the track cones into left and right sides."""
        
        start_rotation += (pi / 2) # 0 rotation in EUFS is offset 900 degrees from our cone positions
        
        # Constants
        MAX_RAY_LENGTH = 10.0
        CONE_RADIUS = 2
        VIEW_FOV_RADIANS = np.radians(110)
        NUMBER_OF_RAYS = 55
        RAY_INCREMENT = 2
        
        rays = {}
        
        # We cast rays across a sector VIEW_FOV_RADIUS wide, centered on the start_rotation
        for ray in np.arange(start_rotation - (VIEW_FOV_RADIANS / 2), start_rotation + (VIEW_FOV_RADIANS / 2), VIEW_FOV_RADIANS / NUMBER_OF_RAYS):
            sin_of_ray = sin(ray) # X component of ray's travel
            cos_of_ray = cos(ray) # Y component of ray's travel
            
            # Then we march the rays until they either "hit" a cone or they reach the MAX_RAY_LENGTH
            for step in np.arange(0.0, MAX_RAY_LENGTH, RAY_INCREMENT):
                new_position = Vector2(start_position.x + (step * sin_of_ray), start_position.y + (step * cos_of_ray))
                if self.intersects_with_track_cones(new_position, CONE_RADIUS):
                    rays[ray] = Vector2.distance(Vector2(0, 0), new_position)
                    break
                
                if step == MAX_RAY_LENGTH - RAY_INCREMENT:
                    rays[ray] = Vector2.distance(Vector2(0, 0), new_position)
           
        # There will often be multiple best-length rays
        longest_ray_length = max(rays.values())
        longest_rays = []
        
        for key, value, in rays.items():
            if value == longest_ray_length:
                longest_rays.append(key)
        
        best_ray = longest_rays[0]
        ideal_ray = start_rotation + (pi / 2)
        
        # Bias our ray selection towards more centered rays
        # This stops the car going to the left or right extremes
        for contender_ray in longest_rays:
            if abs(contender_ray - ideal_ray) < abs(best_ray - ideal_ray):
                best_ray = contender_ray        
        
        line_end_point = Vector2(start_position.x + (sin(best_ray) * MAX_RAY_LENGTH), start_position.y + (cos(best_ray) * MAX_RAY_LENGTH))
        
        return Line(start_position, line_end_point)
    
    def sort_cones(self, dividing_line: Line) -> List[Cone]:
        """Sorts the track's cones into left and right sides."""
        cones = []
        
        for cone in self.track.cones:
            if point_is_on_left_side_of_line(cone.pos, dividing_line):
                cones.append(Cone("blue", cone.pos.x, cone.pos.y))
            else:
                cones.append(Cone("yellow", cone.pos.x, cone.pos.y))
                
        return cones
        

    def solve(self, start_position: Vector2, start_rotation: float, colorblind: bool = True) -> None:
        """Finds a path forward using cone information."""
        self.path = None  # Clear path
        
        t1 = time_ns() / 1000000
                
        if colorblind and self.track is not None:
            self.track.strip_color_information()
            dividing_line = self.get_dividing_line(start_position, start_rotation)
            cones = self.sort_cones(dividing_line)
                
            self.track = Track("tmp", cones)
            
        self.tessellated_triangles = tessellate_track(self.track)
        self.unordered_midpoints = self.get_midpoints(self.tessellated_triangles)
        start_index = self.get_closest_point_index(start_position)
        self.path = self.order_points(start_index)
        self.path.insert(0, start_position)
        
        t2 = time_ns() / 1000000
        
        print("Calculated path in " + str(round(t2 - t1, 2)) + "ms")

    def get_midpoints(self, tessellated_triangles: List[Triangle]) -> List[Vector2]:
        midpoints = []
        lines = []

        for triangle in tessellated_triangles:
            sides = triangle.to_sides()
            lines.append(sides[0])
            lines.append(sides[1])
            lines.append(sides[2])

        for line in lines:
            p1_color = ""
            p2_color = ""

            for cone in self.track.cones:
                if cone.pos == line.p1:
                    p1_color = cone.color
                elif cone.pos == line.p2:
                    p2_color = cone.color

            # Check if p1.color != p2.color
            if p1_color != p2_color:
                midpoints.append(Vector2.midpoint(line.p1, line.p2))

        return midpoints

    def get_closest_point_index(self, position: Vector2) -> int:
        """Gets the array index of the closest midpoint."""
        
        min_distance = float("inf")
        closest_midpoint_index = None

        for i in range(len(self.unordered_midpoints)):
            distance = Vector2.distance(self.unordered_midpoints[i], position)
            if distance < min_distance:
                closest_midpoint_index = i
                min_distance = distance

        return closest_midpoint_index

    def order_points(self, start_point_index: int = 0) -> List[Vector2]:
        """Order points starting from the start point using the nearest neighbor algorithm."""
        
        visited_indices = []

        distances = np.zeros((len(self.unordered_midpoints), len(self.unordered_midpoints)))
        for i in range(len(self.unordered_midpoints)):
            for j in range(len(self.unordered_midpoints)):
                distances[i, j] = Vector2.distance(self.unordered_midpoints[i], self.unordered_midpoints[j])

        current = start_point_index

        while len(visited_indices) < len(self.unordered_midpoints):
            min_distance = float("inf")
            nearest_index = None
            for i in range(len(self.unordered_midpoints)):
                if i not in visited_indices:
                    distance = distances[current, i]
                    if distance < min_distance:
                        min_distance = distance
                        nearest_index = i

            visited_indices.append(nearest_index)
            current = nearest_index

        ordered_points = [self.unordered_midpoints[i] for i in visited_indices]
        return ordered_points
